#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "sort.h"


/******************************************************************************
 *                            CONSTANTS AND LISTS
 *****************************************************************************/

// min/max values of n
const size_t MIN_N = 1000;
const size_t MAX_N = 10000000;

// number of trials to run for each experiment
const size_t NUM_TRIALS = 5;

// total number of sorting algorithms
const size_t NUM_SORTS = 8;

// total number of sorting algorithms considered "slow"
const size_t NUM_SLOW_SORTS = 4;

// don't try "slow" sorts on any n greater than this constant
const size_t SLOW_SORT_THRESHOLD = 100000;

// list of all sorting routines to test
sort all_sorts[] = {
    bubble_sort,
    selection_sort,
    insertion_sort,
    insertion_sort_binary,
    merge_sort,
    merge_sort_swap,
    quick_sort,
    quick_sort_midpoint
};

// labels for sorting routines (order must match all_sorts)
char* sort_labels[] = {
    "Bubble",
    "Selection",
    "Insertion",
    "Insert Binary",
    "Merge",
    "Merge Swap",
    "Quick",
    "Quick Midpoint"
};

// list of sorts that are considered to be "slow" (e.g., n^2)
sort slow_sorts[] = {
    bubble_sort,
    selection_sort,
    insertion_sort,
    insertion_sort_binary
};


/******************************************************************************
 *                           ARRAY HELPER FUNCTIONS
 *****************************************************************************/

employee_t* malloc_employee(int id, int salary)
{
    employee_t *emp = (employee_t*)malloc(sizeof(employee_t));
    if (emp == NULL) {
        printf("Out of memory!\n");
        exit(EXIT_FAILURE);
    }
    emp->id = id;
    emp->salary = salary;
    return emp;
}

void gen_random_employees(employee_t* items[], size_t n)
{
    // generate random salaries
    for (size_t i = 0; i < n; i++) {
        items[i] = malloc_employee(i,
                (int)(((double)rand() / (double)RAND_MAX) * 50000) + 40000);
    }
}

void copy_employees(employee_t* src[], employee_t* dest[], size_t n)
{
    // copy employee pointers from src to dest (shallow copy)
    for (size_t i = 0; i < n; i++) {
        dest[i] = src[i];
    }
}

void print_employees(employee_t* items[], size_t n)
{
    // print an array of employees
    for (size_t i = 0; i < n; i++) {
        if (i > 0) {
            printf(",");
        }
        printf("%dk(%d)", (items[i]->salary/1000), items[i]->id);
    }
    printf("\n");
}

void free_employees(employee_t* items[], size_t n)
{
    // de-allocate memory for an array of employee structs
    for (size_t i = 0; i < n; i++) {
        free(items[i]);
    }
}


/******************************************************************************
 *                           TESTING HELPER FUNCTIONS
 *****************************************************************************/

bool is_slow_sort(sort f)
{
    for (size_t i = 0; i < NUM_SLOW_SORTS; i++) {
        if (slow_sorts[i] == f) {
            return true;
        }
    }
    return false;
}

void do_correctness_tests(size_t n)
{
    // generate some random employees
    data_t *items = (data_t*)malloc(sizeof(data_t) * n);
    gen_random_employees(items, n);
    printf("  %-15s", "Original");
    print_employees(items, n);

    for (size_t i = 0; i < NUM_SORTS; i++) {

        // create a copy of the array to be sorted
        data_t *citems = (data_t*)malloc(sizeof(data_t) * n);
        copy_employees(items, citems, n);

        // run the sort and print the resulting array
        all_sorts[i](citems, n);
        printf("  %-15s", sort_labels[i]);
        print_employees(citems, n);

        // run through results and check for out-of-order values
        bool correct = true;
        for (size_t j = 0; j < n-1; j++) {
            if (cmp(citems[j], citems[j+1]) > 0) {
                correct = false;
            }
        }

        // print error if any out-of-order values were found
        if (!correct) {
            printf("INCORRECT SORT: %s\n", sort_labels[i]);
        }

        free(citems);
    }

    free_employees(items, n);
    free(items);
}

void do_stability_tests(size_t n)
{
    // generate some random employees, ensuring (using the pidgeonhole
    // principle) that there will be at least two employees with the same salary
    data_t *items = (data_t*)malloc(sizeof(data_t) * n);
    for (size_t i = 0; i < n; i++) {
        items[i] = malloc_employee(i, 40000 + 1000*(rand()%(n/2)));
    }
    /*printf("  %-15s" "Original");*/
    /*print_employees(items, n);*/

    for (size_t i = 0; i < NUM_SORTS; i++) {

        // create a copy of the array to be sorted
        data_t *citems = (data_t*)malloc(sizeof(data_t) * n);
        copy_employees(items, citems, n);

        // run the sort
        all_sorts[i](citems, n);

        // run through results and check for out-of-order values
        bool stable = true;
        for (size_t j = 0; j < n-1; j++) {
            if (cmp(citems[j], citems[j+1]) == 0 &&
                    citems[j]->id > citems[j+1]->id) {
                /*printf("  Instability detected at index %lu\n", j);*/
                stable = false;
            }
        }

        // print stability result
        printf("  %-15s %s\n", sort_labels[i], stable ? "stable" : "NOT STABLE");
        /*print_employees(citems, n);*/

        free(citems);
    }

    free_employees(items, n);
    free(items);
}

void do_performance_test(data_t items[], size_t n, sort f, const char *label)
{
    clock_t begin, end;
    double sum_time = 0.0;

    // run several trials and average the results
    for (size_t i = 0; i < NUM_TRIALS; i++) {

        // create a copy of the array to be sorted
        data_t *citems = (data_t*)malloc(sizeof(data_t) * n);
        copy_employees(items, citems, n);

        // run the sort and time it
        begin = clock();
        f(citems, n);
        end = clock();
        sum_time += ((double)(end-begin)/(double)CLOCKS_PER_SEC);

        free(citems);
    }

    // print results
    printf("  %-15s %12lu %12.4f\n", label, n, (sum_time / NUM_TRIALS));
}

void do_all_performance_tests(data_t items[], size_t n)
{
    // run every sort unless it's a slow sort and n is too big
    for (size_t i = 0; i < NUM_SORTS; i++) {
        if (!is_slow_sort(all_sorts[i]) || n < SLOW_SORT_THRESHOLD) {
            do_performance_test(items, n, all_sorts[i], sort_labels[i]);
        }
    }
}

#define SWAP(A,I,J) data_t tmp = (A)[(I)]; (A)[(I)] = (A)[(J)]; (A)[(J)] = tmp;

void do_inorder_performance_tests(size_t n)
{
    // generate input with numbers in order (already sorted)
    data_t *items = (data_t*)malloc(sizeof(data_t) * n);
    for (size_t i = 0; i < n; i++) {
        items[i] = malloc_employee(i, 40000.0 + i);
    }
    // randomly permute a few values from alternate sides of the array
    for (size_t i = 0; i < n/100; i++) {
        size_t j = rand() % (n/2);
        size_t k = rand() % (n/2) + n/2;
        SWAP(items, j, k);
    }
    do_all_performance_tests(items, n);
    free_employees(items, n);
    free(items);
}

void do_random_performance_tests(size_t n)
{
    // generate input with random numbers drawn from the range [0, n)
    data_t *items = (data_t*)malloc(sizeof(data_t) * n);
    gen_random_employees(items, n);
    do_all_performance_tests(items, n);
    free_employees(items, n);
    free(items);
}


/******************************************************************************
 *                            PROGRAM ENTRY POINT
 *****************************************************************************/

int main()
{
    // seed random number generator
    srand((unsigned)time(NULL));

    // correctness and stability tests
    printf("Correctness:\n");
    do_correctness_tests(10);
    printf("\nStability:\n");
    do_stability_tests(10);

    // performance tests
    for (size_t n = MIN_N; n < MAX_N+1; n *= 10) {
        printf("\nRandom-order:\n");
        do_random_performance_tests(n);
        printf("\nNearly-sorted:\n");
        do_inorder_performance_tests(n);
    }
}

